/*----------------------------------------------------------------------
  File    : cluster4.c
  Contents: cluster initialization and exact hard c-means algorithms
  Author  : Christian Borgelt
  History : 2018.06.10 file created (as part of file cli.c)
            2018.06.10 other initialization methods added (-2018.06.20)
            2018.07.12 simplified AFK-MC2 added (stoch. univ. sampling)
            2018.07.17 (bounds-based) exact hard c-means methods added
            2018.07.19 initialization and update functions moved here
            2018.07.25 trimmed maximin init. added (cut top quantile)
            2018.07.26 hybrid cluster update/optimization method added
            2018.08.17 restricted to data matrices (no more tables)
            2018.08.27 annular algorithm [Drake 2013] added
            2018.08.28 yinyang algorithm [Ding et al. 2015] added
            2018.08.29 stratum algorithm (annular variant) added
            2018.08.31 shallot algorithm (improved exponion) added
            2018.09.01 coordinate grid (stratum) algorithm added
            2018.09.03 sorted lists instead of heap for hybrid algorithm
            2019.02.12 bug in call to quantile() fixed (argument k)
            2019.07.09 distribution based initialization added
            2019.07.10 non-squared distance based kmeans++ init. added
            2019.11.15 adapted to changed interface of mat_shuffle()
----------------------------------------------------------------------*/
#include <stdio.h>
#include <float.h>
#include <assert.h>
#include "arrays.h"
#include "random.h"

#ifndef MAT_RDWR
#define MAT_RDWR
#endif

#include "cluster.h"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#ifdef BENCH                    /* if benchmark version */
#define COUNTDC(n)  distcnt += (size_t)(n)
#else                           /* count distance computations */
#define COUNTDC(n)              /* if standard version */
#endif                          /* do not count distance computations */

#define SWAP(a,b,x) do { x = a; a = b; b = x; } while (0)

#define SORTED      1           /* whether to use sorted lists */
#define THRESH      5           /* sorted list threshold in hybrid */

/*--------------------------------------------------------------------*/

#define int         1           /* to check definitions */
#define long        2           /* for certain types */
#define ptrdiff_t   3

/*--------------------------------------------------------------------*/

#if   DIMID==int
#ifndef dim_shuffle
#define dim_shuffle(a,n,r)        int_shuffle(a,n,r)
#endif
#ifndef dim_qsort
#define dim_qsort(i,n,d)          int_qsort(i,n,d)
#endif
#ifndef dim_isort
#define dim_isort(i,n,d,a)        i2d_qsort(i,n,d,a)
#endif
#ifndef dim_ibisect
#define dim_ibisect(i,n,d,a)      ((DIMID)i2d_bisect(i,n,d,a))
#endif
#ifndef dim_psort
#define dim_psort(i,n,d,a,c,x)    i2p_qsort(i,n,d,a,c,x)
#endif

#elif DIMID==long
#ifndef dim_shuffle
#define dim_shuffle(a,n,r)        lng_shuffle(a,n,r)
#endif
#ifndef dim_qsort
#define dim_qsort(i,n,d)          lng_qsort(i,n,d)
#endif
#ifndef dim_isort
#define dim_isort(i,n,d,a)        l2d_qsort(i,n,d,a)
#endif
#ifndef dim_ibisect
#define dim_ibisect(i,n,d,a)      ((DIMID)l2d_bisect(i,n,d,a))
#endif
#ifndef dim_psort
#define dim_psort(i,n,d,a,c,x)    l2p_qsort(i,n,d,a,c,x)
#endif

#elif DIMID==ptrdiff_t
#ifndef dim_shuffle
#define dim_shuffle(a,n,r)        dif_shuffle(a,n,r)
#endif
#ifndef dim_qsort
#define dim_qsort(i,n,d)          dif_qsort(i,n,d)
#endif
#ifndef dim_isort
#define dim_isort(i,n,d,a)        x2d_qsort(i,n,d,a)
#endif
#ifndef dim_ibisect
#define dim_ibisect(i,n,d,a)      ((DIMID)x2d_bisect(i,n,d,a))
#endif
#ifndef dim_psort
#define dim_psort(i,n,d,a,c,x)    x2p_qsort(i,n,d,a,c,x)
#endif

#else
#error "DIMID must be either 'int', 'long' or 'ptrdiff_t'"
#endif

/*--------------------------------------------------------------------*/

#undef int                      /* remove preprocessor definitions */
#undef long                     /* needed for the type checking */
#undef ptrdiff_t

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef double DISTFN (const double *a, const double *b, DIMID n);

/*----------------------------------------------------------------------
  Cluster Initialization from Data Points
----------------------------------------------------------------------*/

static double quantile (DIMID *index, DIMID n, DIMID k,
                        const double *array)
{                               /* --- collect values above quantile */
  DIMID  *l, *r, *q;            /* pointers to exchange positions */
  DIMID  t;                     /* pivot element and exchange buffer */
  double p;                     /* buffer for pivot element */

  assert(index && array         /* check the function arguments */
  &&    (n > 0) && (k >= 0) && (k < n));
  q = index +k;                 /* get the quantile location */
  while (n > 1) {               /* sections split loop (quicksort) */
    l = index; r = index +n-1;  /* start at left and right boundary */
    if (array[*l] < array[*r]){ /* get the first and last elements */
      t = *l; *l = *r; *r = t; }/* and bring them into right order */
    p = array[index[n/2]];      /* get the middle element as pivot */
    if      (p > array[*l])     /* compute median of three */
      p = array[*l];            /* (first, middle, last) */
    else if (p < array[*r])     /* to find a better pivot */
      p = array[*r];            /* for the split/exchanges */
    while (1) {                 /* split and exchange loop */
      while (array[*++l] > p);  /* skip greater elems. on the left */
      while (array[*--r] < p);  /* skip smaller elems. on the right */
      if (l >= r) break;        /* if at most one element left, abort */
      t = *l; *l = *r; *r = t;  /* otherwise exchange elements */
    }
    if (l <= r) {               /* if exactly one element left */
      if (l == q) return array[*q];
      l++; r--;                 /* if at quantile, return quantile, */
    }                           /* otherwise skip the pivot element */
    if (q <= r)                 /* if quantile in left/lower section */
      n = 1 +(DIMID)(r-index);  /* compute the new section size */
    else {                      /* if quantile in right/upper section */
      n = n -(DIMID)(l-index);  /* compute the new section size */
      index = l;                /* section now starts with l */
    }                           /* (skip the left section) */
  }
  return array[*q];             /* return the quantile element */
}  /* quantile() */

/*--------------------------------------------------------------------*/

static void sift (DIMID *heap, DIMID lft, DIMID rgt, DIMID idx,
                  const double *dsts)
{                               /* --- let index sift down in heap */
  DIMID  i = lft +lft +1;       /* compute index of first successor */
  double d = dsts[idx];         /* distance of data point to let sift */

  while (i < rgt) {             /* sift loop (while within heap) */
    if ((i+1 < rgt) && (dsts[heap[i]] > dsts[heap[i+1]]))
      i++;                      /* if second successor is less */
    if (d <= dsts[heap[i]])     /* if the successor is less */
      break;                    /* than the sift element, */
    heap[lft] = heap[i];        /* let the successor ascend in heap */
    lft = i; i += i+1;          /* compute index of first successor */
  }
  heap[lft] = idx;              /* store the sift element */
}  /* sift() */

/*--------------------------------------------------------------------*/

int cls_ctrsel (MATRIX *ctrs, MATRIX *data, int method,
                double size, double trim, DIMID *outids, DIMID outcnt,
                RANDFN randfn)
{                               /* --- init. clusters w/ data points */
  DIMID  n, i, j, x, y;         /* number of data points, point index */
  DIMID  c, k;                  /* number of clusters, cluster index */
  DIMID  m;                     /* number of dimensions/vector elems. */
  DIMID  s, t;                  /* section and trim size, seq. length */
  double *mem;                  /* work memory (one memory block) */
  double *cps;                  /* cumulated probabilities for afkmc2 */
  double *dsts;                 /* (minimum) distances to centers */
  double d, dmax;               /* current and maximum distance */
  double dsum;                  /* current sum of (min.) distances */
  double delta;                 /* distance step width for sampling */
  DIMID  *hcis;                 /* highest used cluster center id */
  DIMID  *s4mc;                 /* samples for Markov chain */
  DIMID  *heap;                 /* heap for quantiled kmeans++ */
  double *vec = NULL;           /* current data point/cluster center */
  DISTFN *distfn;               /* distance function for kmeans++ */
  #ifdef BENCH                  /* if benchmark version */
  size_t distcnt = 0;           /* number of distance computations */
  #endif                        /* (incremented for each cls_sqrdist) */

  assert(ctrs && data && randfn /* check the function arguments */
  &&    (mat_colcnt(ctrs) == mat_colcnt(data))
  &&    (mat_rowcnt(ctrs) > 0) && (mat_rowcnt(data) > 0)
  &&    (!outids || (outcnt <= 0)));
  n = mat_rowcnt(data);         /* get the number of data points */
  c = mat_rowcnt(ctrs);         /* and the number of clusters */
  m = mat_colcnt(ctrs);         /* and the number of dimensions */
  assert(outcnt <= n);          /* check the number of outliers */

  if (method < CLS_POINTS) {    /* distribution based initialization */
    mem = (double*)malloc((size_t)(m+m) *sizeof(double));
    if (!mem) return -1;        /* create minimum/maximum vectors */
    cps = mem +m;               /* to find min. hyperbox around data */
    for (j = 0; j < m; j++) {   /* initialize coords. to +/-infinity */
      mem[j] = INFINITY; cps[j] = -INFINITY; }
    for (i = 0; i < n; i++) {   /* traverse the data points */
      vec = mat_row(data, i);   /* get the next data point */
      for (j = 0; j < m; j++) { /* traverse the dimensions */
        if (vec[j] < mem[j]) mem[j] = vec[j];
        if (vec[j] > cps[j]) cps[j] = vec[j];
      }                         /* find minimum and maximum values */
    }                           /* for all coordinates */
    if (method == CLS_CENTER) { /* if to init. to data space center */
      for (j = 0; j < m; j++)   /* compute the center coordinates */
        mem[j] = 0.5 *(mem[j] +cps[j]);
      for (k = 0; k < c; k++)   /* set all cluster centers */
        mat_rowset(ctrs, k, mem);
    }
    if (method == CLS_UNIFORM){ /* if uniform sample from d.s. */
      for (j = 0; j < m; j++)   /* compute the coordinate ranges */
        cps[j] -= mem[j];       /* (difference between max. and min.) */
      for (k = 0; k < c; k++) { /* traverse the cluster centers */
        vec = mat_row(ctrs, k); /* get the next cluster center */
        for (j = 0; j < m; j++) /* traverse the dimensions */
          vec[j] = mem[j] +randfn() *cps[j];
      }                         /* init. to uniformly drawn point */
    }
    if ((method == CLS_DIAG)    /* if diagonal of data space */
    ||  (method == CLS_LATIN)){ /* or latin hypercube sampling */
      for (j = 0; j < m; j++) { /* traverse the dimensions */
        delta = (cps[j] -mem[j]) /(double)c;      /* get offset */
        d     = mem[j] +0.5 *delta;     /* and coordinate delta */
        for (k = 0; k < c; k++){/* traverse the cluster centers */
          mat_set(ctrs, k, j, d); d += delta; }
      }                         /* set equidistant coordinates */
    }                           /* for all dimensions */
    if (method == CLS_LATIN){   /* if latin hypercube sampling */
      for (k = c; --k > 0; ) {  /* shuffling needs c-1 steps */
        for (j = 0; j < m; j++) { /* traverse the dimensions */
          x = (DIMID)((double)(k+1) *randfn());
          if      (x > k) x = k;  /* compute a random index in */
          else if (x < 0) x = 0;  /* the remaining set of centers */
          d = mat_get(ctrs, k, j);
          mat_set(ctrs, k, j, mat_get(ctrs, x, j));
          mat_set(ctrs, x, j, d);
        }                       /* exchange j-th coordinate of */
      }                         /* k-th and x-th cluster center */
    }                           /* (shuffle dimensions independently) */
    free(mem);                  /* free the working memory */
    #ifdef BENCH                /* if benchmark version */
    fprintf(stderr, "dc: 0 ");  /* print number of distance comps. */
    #endif
    return 0;                   /* abort the function */
  }

  if (outids) {                 /* if outlier indicators are given */
    for (i = 0; i < outcnt; i++)
      mat_rowexg(data, outids[i], data, --n);
  }                             /* move outliers to the end */
  mat_shuffle(data,randfn,0,n); /* shuffle the data points */

  if (method == CLS_POINTS) {   /* if to sample centers uniformly, */
    for (k = 0; k < c; k++)     /* select the first c rows/tuples */
      mat_rowput(ctrs, k, mat_row(data, k % n));
    #ifdef BENCH                /* if benchmark version */
    fprintf(stderr, "dc: 0 ");  /* print number of distance comps. */
    #endif
    return 0;                   /* sampling uniformly at random */
  }                             /* is complete, so abort function */

  if (method < CLS_RKMPP)       /* if standard methods, */
    distfn = vec_sqrdist;       /* use the squared distance */
  else {                        /* if alternative methods, */
    distfn = vec_dist;          /* use the normal distance */
    method -= (CLS_RKMPP -CLS_KMPP);
  }                             /* replace the method identifier */
  s = (size > 0) ? (DIMID)(size *(double)n)
    : (size < 0) ? (DIMID)-size : n;
  if (s >  n) s = n;            /* get size of section/chain length */
  if (s <= 0) s = 1;            /* and ensure a valid value */
  t = (trim > 0) ? (DIMID)(trim *(double)n)
                 : (DIMID)-trim;
  if (t > 0) t += 1;            /* get number of distances to trim */
  if (t > s) t  = s;            /* and ensure a valid value */
  x = ((method == CLS_AFKMC2)   /* size of additional arrays */
    || (method == CLS_SAFKMC2)) ? n+1 : 0;
  y = ((method == CLS_QKMPP)    /* (cum. probs or heap/chain elems.) */
    || (method == CLS_SAFKMC2)) ? s
    : ((method == CLS_MAXIMIN) && (t > 0)) ? t : 0;
  mem = (double*)malloc((size_t)(x+n+1) *sizeof(double)
                       +(size_t)(n+y)   *sizeof(DIMID));
  if (!mem) return -1;          /* allocate work memory for */
  cps  = mem +1;                /* the cumulated probabilities */
  dsts = mem +x;                /* and (minimal) distance array, */
  hcis = (DIMID*)(dsts +n+1);   /* highest used cluster index array */
  s4mc = hcis +n;               /* and array of data point samples */
  heap = s4mc;                  /* which may also be a heap */
  mat_rowput(ctrs, 0, vec = mat_row(data, 0));
  dsts[0] = 0.0;                /* set first data point as cluster */
  /* Since the data points (rows/tuples) were shuffled before,   */
  /* the first point is a data point chosen uniformly at random. */
  for (i = 0; ++i < n; )        /* compute and store distances */
    dsts[i] = distfn(mat_row(data, i), vec, m);
  COUNTDC(n-1);                 /* count the computed distances */

  if ((method == CLS_MAXIMIN)   /* if maximum minimum distance sel. */
  &&  (size == 0)) {            /* and to skip initial data point */
    if (t <= 0) {               /* if strict maximin (no trimming) */
      for (d = 0, i = k = 0; i < n; i++) {
        if (dsts[i] > d)        /* find index of data point that */
          d = dsts[k = i];      /* is farthest from initial point */
      } }                       /* (maximum distance point) */
    else {                      /* if trimmed maximin (trim top t) */
      dsts[heap[0] = n] = -1.0; /* set a heap sentinel */
      for (j = t, i = 0; i < n; i++) {
        if (dsts[i] <= dsts[heap[0]])
          continue;             /* if less than heap data, skip point */
        if (j > 0) j -= 1;      /* get (next) position in heap */
        sift(heap, j, t, i, dsts);
      }                         /* store data point (index) in heap */
      k = heap[0];              /* get the closest data point */
    }                           /* in the heap of farthest points */
    mat_rowput(ctrs, 0, vec = mat_row(data, k));
    dsts[k] = 0.0;              /* set found point as first center */
    for (i = 0; i < n; i++)     /* traverse remaining data points */
      if (i != k) dsts[i] = vec_sqrdist(mat_row(data, i), vec, m);
  }                             /* compute distances to chosen center */

  if ((method == CLS_MAXIMIN)   /* if maximum minimum distance based */
  &&  (t <= 0)) {               /* data point selection (untrimmed) */
    memset(hcis, 0, (size_t)n *sizeof(DIMID));
    for (i = k = 0; ++k < c; ){ /* select remaining cluster centers */
      x = 0; dmax = 0.0;        /* init. index and maximal distance */
      for (j = 0; j < s; j++) { /* traverse the data point section */
        i = (i+1) % n;          /* get the next data point index */
        if (dsts[i] <= dmax)    /* if smaller than current maximum, */
          continue;             /* data point will not be selected */
        vec = mat_row(data, i); /* get the data point to process */
        while (hcis[i] < k-1) { /* traverse (new) cluster centers */
          d = vec_sqrdist(mat_row(ctrs, ++hcis[i]), vec, m);
          COUNTDC(1);           /* compute distance to new clusters */
          if ((d < dsts[i]) && ((dsts[i] = d) < dmax))
            break;              /* find minimal distance to cluster */
        }                       /* and update highest cluster index */
        if (dsts[i] > dmax) dmax = dsts[x = i];
      }                         /* update data point with max. dist. */
      dsts[x] = 0.0;            /* mark data point as selected */
      mat_rowput(ctrs, k, mat_row(data, x));
    } }                         /* set the next cluster center */
  else if (method == CLS_MAXIMIN) { /* if trimmed maximin selection */
    memset(hcis, 0, (size_t)n *sizeof(DIMID));
    dsts[n] = -1.0;             /* init. highest used cluster ids */
    for (k = 0; ++k < c; ) {    /* select remaining cluster centers */
      heap[0] = n;              /* set a heap sentinel */
      for (j = t, i = 0; i < n; i++) {
        if (dsts[i] <= dsts[heap[0]])
          continue;             /* if less than heap data, skip point */
        vec = mat_row(data, i); /* get the data point to process */
        while (hcis[i] < k-1) { /* traverse (new) cluster centers */
          d = vec_sqrdist(mat_row(ctrs, ++hcis[i]), vec, m);
          COUNTDC(1);           /* compute distance to new clusters */
          if (d < dsts[i]) dsts[i] = d;
        }                       /* find minimal distance to cluster */
        if (dsts[i] <= dsts[heap[0]])
          continue;             /* if less than heap data, skip point */
        if (j > 0) j -= 1;      /* get (next) position in heap */
        sift(heap, j, t, i, dsts);
      }                         /* store data point (index) in heap */
      x = heap[0];              /* get index of maximum trimmed dist. */
      if (x >= n) x = n-1;      /* ensure a proper data point index */
      dsts[x] = 0.0;            /* mark the data point as selected */
      mat_rowput(ctrs, k, mat_row(data, x));
    } }                         /* set the next cluster center */
  else if (method == CLS_KMPP){ /* if kmeans++ initialization */
    for (i = 0; i < n; i++)     /* initialize the cluster indices */
      hcis[i] = i;              /* for the quantile computation */
    for (k = 0; ++k < c; ) {    /* select remaining cluster centers */
      if (k > 1) {              /* compute distance to new centers */
        vec = mat_row(ctrs,k-1);/* get last selected cluster center */
        for (i = 0; i < n; i++){/* traverse unselected data points */
          hcis[i] = i;          /* reinitialize the point indices */
          /* This reinitialization of the point indices is necessary, */
          /* because a preceding quantile finding leaves this array   */
          /* in a state that renders subsequent quantile finding slow.*/
          if (dsts[i] <= 0.0) continue;
          d = distfn(mat_row(data, i), vec, m);
          COUNTDC(1);           /* compute distance to new cluster */
          if (d < dsts[i]) dsts[i] = d;
        }                       /* update the closest distance */
      }                         /* to a cluster if necessary */
      if (s < n)                   /* if to use only a data subset, */
        quantile(hcis,n,s-1,dsts); /* compute req. lower quantile */
      if (t > 0)                   /* if to trim uppermost quantile */
        quantile(hcis,s,t-1,dsts); /* compute req. upper quantile */
      if ((s < n) || (t > 0)) { /* if to restrict to quantile range */
        for (dsum = 0, i = t; i < s; i++)
          dsum += dsts[hcis[i]];/* sum the (squared) distances */
        dmax = dsum *randfn();  /* choose a random distance sum */
        for (d = 0, i = t; i < n; i++) {
          if ((d += dsts[hcis[i]]) >= dmax)
            break;              /* find the selected data point */
        }                       /* in the requested quantile */
        if (i >= n) i = n-1;    /* ensure a valid data point index */
        i = hcis[i]; }          /* get index of the chosen center */
      else {                    /* if to use the whole data set */
        for (dsum = 0, i = 0; i < n; i++)
          dsum += dsts[i];      /* sum the (squared) distances */
        dmax = dsum *randfn();  /* choose a random distance sum */
        for (d = 0, i = t; i < n; i++)
          if ((d += dsts[i]) >= dmax)
            break;              /* find the selected data point */
        if (i >= n) i = n-1;    /* ensure a proper data point index */
      }
      dsts[i] = 0.0;            /* mark the data point as selected */
      mat_rowput(ctrs, k, mat_row(data, i));
    } }
  else if (method == CLS_QKMPP){/* if quantiled kmeans++ init. */
    memset(hcis, 0, (size_t)n *sizeof(DIMID));
    dsts[n] = -1.0;             /* init. highest used cluster ids */
    for (k = 0; ++k < c; ) {    /* select remaining cluster centers */
      heap[0] = n;              /* set a heap sentinel */
      for (j = s, i = 0; i < n; i++) {
        if (dsts[i] <= dsts[heap[0]])
          continue;             /* if less than heap data, skip point */
        vec = mat_row(data, i); /* get the data point to process */
        while (hcis[i] < k-1) { /* traverse (new) cluster centers */
          d = distfn(mat_row(ctrs, ++hcis[i]), vec, m);
          COUNTDC(1);           /* compute distance to new clusters */
          if (d < dsts[i]) dsts[i] = d;
        }                       /* find minimal distance to cluster */
        if (dsts[i] <= dsts[heap[0]])
          continue;             /* if less than heap data, skip point */
        if (j > 0) j -= 1;      /* get (next) position in heap */
        sift(heap, j, s, i, dsts);
      }                         /* store data point (index) in heap */
      if (t > 0)                /* if to trim uppermost quantile, */
        quantile(heap,s,t-1,dsts); /* compute req. upper quantile */
      for (dsum = 0, i = t; i < s; i++)
        dsum += dsts[heap[i]];  /* sum the (squared) distances */
      dmax = dsum *randfn();    /* choose a random distance sum */
      for (d = 0, i = t; i < s; i++) {
        if ((d += dsts[heap[i]]) >= dmax)
          break;                /* find the selected data point */
      }                         /* in the requested quantile */
      if ((i >= s) && (s > 0)) i = s-1;
      dsts[x = heap[i]] = 0.0;  /* mark the data point as selected */
      mat_rowput(ctrs, k, mat_row(data, x));
    } }                         /* set the next cluster center */
  else if ((method == CLS_AFKMC2)  /* if assump.-free Markov chain */
  ||       (method == CLS_SAFKMC2)) {               /* Monte Carlo */
    memset(hcis, 0, (size_t)n *sizeof(DIMID));
    cps[-1] = 0.0;              /* sentinel for cumulated probs. */
    cps[ 0] = d = 0.5/(double)n;/* compute uniform part and norm. */
    for (dsum = 0, i = 0; i < n; i++)
      dsum += dsts[i];          /* sum the (squared) distances */
    dsum = (dsum > 0) ? 0.5/dsum : 0;
    for (i = 0; ++i < n; )      /* compute selection probabilities */
      cps[i] = dsts[i]*dsum +d; /* distance part + uniform part */
    for (dsum = 0, i = 0; i < n; i++)
      cps[i] = dsum += cps[i];  /* compute cumulated probabilities */
    for (k = 0; ++k < c; ) {    /* select remaining cluster centers */
      if (method == CLS_AFKMC2){/* if standard version */
        x = (DIMID)dbl_bisect(randfn(), cps, (size_t)n);
        if (x >= n) x = n-1; }  /* sample accord. to d^2 distribution */
      else {                    /* if simplified version */
        delta = dsum /s;        /* get increment and offset for */
        d     = randfn()*delta; /* stochastic universal sampling */
        for (dmax = 0, i = j = 0; i < s; i++) {
          while ((j < n) && (d > cps[j]))
            j += 1;             /* find selected data point */
          if (j >= n) j = n-1;  /* based on next marker position */
          s4mc[i] = j;          /* store data point in sample */
          d += delta;           /* go to the next marker position */
        }                       /* (stochastic universal sampling) */
        dim_shuffle(s4mc, (size_t)s, randfn);
        x = s4mc[0];            /* shuffle the chosen samples and */
      }                         /* use first point as chain start */
      for (i = 0; (dsts[x] == 0.0) && (i < n); i++)
        x = (x+1) % n;          /* ensure point is not cluster center */
      vec = mat_row(data, x);   /* get the data point to process */
      while (hcis[x] < k-1) {   /* traverse (new) cluster centers */
        d = distfn(mat_row(ctrs, ++hcis[x]), vec, m);
        COUNTDC(1);             /* compute distance to new clusters */
        if (d < dsts[x]) dsts[x] = d;
      }                         /* find minimal distance to cluster */
      for (i = 0; ++i < s; ) {  /* execute the Markov chain */
        if (method == CLS_SAFKMC2) /* if simplified version, */
          y = s4mc[i];          /* get the next sample point */
        else {                  /* if standard version */
          y = (DIMID)dbl_bisect(randfn(), cps, (size_t)n);
          if (y >= n) y = n-1;  /* choose a random data point */
        }                       /* according to d^2 distribution */
        for (j = 0; (dsts[y] == 0.0) && (j < n); j++)
          y = (y+1) % n;        /* ensure point is not cluster center */
        vec = mat_row(data, y); /* get the data point to process */
        while (hcis[y] < k-1) { /* traverse (new) cluster centers */
          d = distfn(mat_row(ctrs, ++hcis[y]), vec, m);
          COUNTDC(1);           /* compute distance to new clusters */
          if (d < dsts[y]) dsts[y] = d;
        }                       /* find minimal distance to cluster */
        if (dsts[y] *(cps[x] -cps[x-1])
        >   dsts[x] *(cps[y] -cps[y-1]) *randfn())
          x = y;                /* accept new data point */
      }                         /* based on random condition */
      dsts[x] = 0.0;            /* mark data point as selected */
      mat_rowput(ctrs, k, mat_row(data, x));
    }                           /* set the selected data point */
  }
  #ifdef BENCH                  /* if benchmark version */
  fprintf(stderr, "dc: %zd ", distcnt);
  #endif                        /* print number of distance comps. */
  free(mem);                    /* delete the work memory */
  return 0;                     /* return 'ok' */
}  /* cls_ctrsel() */

/*----------------------------------------------------------------------
  References for cluster center initialization methods:
  [Arthur and Vassilvitskii 2007] (kmeans++)
    David Arthur and Sergei Vassilvitskii.
    k-means++: The Advantages of Careful Seeding.
    Proc. 18th Ann. ACM-SIAM Symp. on Discrete Algorithms
    (SODA 2007, New Orleans, LA), 1027–1035.
    Society for Industrial and Applied Mathematics Philadelphia
    (SIAM), PA, USA 2007
    doi:10.1145/1283383.1283494
  [Hathaway et al. 2006] (maximin)
    Richard J. Hathaway, James C. Bezdek, and Jacalyn M. Huband.
    Maximin Initialization for Cluster Analysis.
    Proc. 11th Iberoamerican Cong. in Pattern Recognition
    (CIARP 2006, Cancun, Mexico), 14-26.
    Springer, New York, NY, USA 2006
    doi:10.1007/11892755_2
  [Eglajs and Audze 1977] (latin hypercube sampling)
    V. Eglajs and P. Audze.
    New Approach to the Design of Multifactor Experiments.
    Problems of Dynamics and Strengths 35:104-107.
    Zinatne Publishing House, Riga, Latvia 1977
  [McKay et al. 1979]
    M.D. McKay, R.J. Beckman, and W.J. Conover.
    A Comparison of Three Methods for Selecting Values
    of Input Variables in the Analysis of Output from a Computer Code.
    Technometrics 21(2):239–245.
    American Statistical Association, Alexandria, VA, USA 1979
    doi:10.2307/1268522
----------------------------------------------------------------------*/

void cls_ctrjtr (MATRIX *ctrs, double range, RANDFN randfn)
{                               /* --- jitter cluster centers */
  DIMID  c, k;                  /* number of clusters, cluster index */
  DIMID  m, j;                  /* number of dimensions/vector elems. */
  double *vec;                  /* to traverse the cluster centers */

  assert(ctrs && randfn         /* check the function arguments */
  &&    (mat_rowcnt(ctrs) > 0));
  if (range <= 0) return;       /* if no range is given, abort */
  c = mat_rowcnt(ctrs);         /* and the number of clusters */
  m = mat_colcnt(ctrs);         /* and the number of dimensions */
  for (k = 0; k < c; k++) {     /* traverse the cluster centers */
    vec = mat_row(ctrs, k);     /* get the next cluster center */
    for (j = 0; j < m; j++)     /* traverse the dimensions */
      vec[j] += (2 *randfn() -1) *range;
  }                             /* add a random offset to all values */
}  /* cls_ctrjtr() */

/*----------------------------------------------------------------------
  (Bounds-based) Exact Hard c-Means
----------------------------------------------------------------------*/

static int dimcmp (const void *p1, const void *p2, void *data)
{                               /* --- compare by some dimension */
  DIMID  col = (DIMID)(size_t)data;
  double v1  = ((double*)p1)[col];
  double v2  = ((double*)p2)[col];
  return (v1 < v2) ? -1 : (v1 > v2) ? +1 : 0;
}  /* dimcmp() */               /* return difference of column values */

/*--------------------------------------------------------------------*/

static DIMID bisect (double key, const DIMID *index, DIMID n,
                     double **array, DIMID col)
{                               /* --- do a bisection search */
  DIMID  l, r, m;               /* array indices */
  double t;                     /* array element */

  assert(array                  /* check the function arguments */
  &&    (n > 0) && (col >= 0));
  for (l = 0, r = n; l < r; ) { /* while search range is not empty */
    t = array[index[m = (l+r)/2]][col]; /* compare given key */
    if      (key > t) l = m+1;  /* to the middle element and */
    else if (key < t) r = m;    /* adapt the search range */
    else return m;              /* according to the result */
  }                             /* if match found, return index */
  return l;                     /* return the insertion position */
}  /* bisect() */

/*--------------------------------------------------------------------*/

static void xsift (double *dsts, DIMID *cids,
                   DIMID l, DIMID r, double d, DIMID t)
{                               /* --- let index sift down in heap */
  DIMID i = l +l +1;            /* compute index of first successor */
  while (i < r) {               /* sift loop (while within heap) */
    if ((i+1 < r) && (dsts[i] < dsts[i+1]))
      i++;                      /* if second successor is greater */
    if (d >= dsts[i])           /* if the successor is greater */
      break;                    /* than the sift element, */
    dsts[l] = dsts[i];          /* let the successor ascend in heap */
    cids[l] = cids[i];          /* (both distance and cluster index) */
    l = i; i += i+1;            /* compute index of first successor */
  }
  dsts[l] = d;                  /* store the sift element */
  cids[l] = t;                  /* (both distance and cluster index) */
}  /* xsift() */

/*--------------------------------------------------------------------*/

static double* c2cdist (MATRIX *ctrs, double *dsts, int mode)
{                               /* --- cluster to cluster dists. */
  DIMID   i, k;                 /* loop variables */
  DIMID   c;                    /* number of clusters */
  DIMID   m;                    /* number of (input) dimensions */
  size_t  z;                    /* size of memory block to allocate */
  double  *vec;                 /* to traverse the centers */
  double  d, *p;                /* cluster to cluster distance */

  assert(ctrs);                 /* check the function arguments */
  c = mat_rowcnt(ctrs);         /* get the number of clusters and */
  m = mat_colcnt(ctrs);         /* the number of (input) dimensions */
  if (!dsts) {                  /* if no distance array is given, */
    z = (mode <= 0) ? (size_t) c/* compute size of memory block */
      : (mode == 1) ? (size_t)(c +c*(c-1)/2)
      :               (size_t)(c +c*c);
    dsts = (double*)malloc(z *sizeof(double));
    if (!dsts) return NULL;     /* create a distance array */
  }                             /* (mode-dependent shape) */
  for (i = 0; i < c; i++)       /* initialize the distances */
    dsts[i] = DBL_MAX;          /* to the closest other cluster */
  if (mode == 0) {              /* if only minimal distances */
    for (i = 0; ++i < c; ) {    /* traverse the clusters */
      vec = mat_row(ctrs, i);   /* get the cluster center */
      for (k = 0; k < i; k++) { /* traverse the cluster pairs */
        d = vec_dist(vec, mat_row(ctrs, k), m);
        if (d < dsts[i]) dsts[i] = d;
        if (d < dsts[k]) dsts[k] = d;
      }                         /* compute distance between clusters, */
    } }                         /* update closest cluster distance */
  else if (mode == 1) {         /* if also triangular distance matrix */
    p = dsts +c;                /* get triangular distance matrix */
    for (i = 0; ++i < c; ) {    /* traverse the clusters */
      vec = mat_row(ctrs, i);   /* get the cluster center */
      for (k = 0; k < i; k++) { /* traverse the cluster pairs */
        *p++ = d = vec_dist(vec, mat_row(ctrs, k), m);
        if (d < dsts[i]) dsts[i] = d;
        if (d < dsts[k]) dsts[k] = d;
      }                         /* compute distance between clusters, */
    } }                         /* update closest cluster distance */
  else {                        /* if also square distance matrix */
    p = dsts +c; p[0] = 0;      /* get triangular distance matrix */
    for (i = 0; ++i < c; ) {    /* traverse the clusters */
      vec = mat_row(ctrs, i);   /* get the cluster center */
      p[i*c+i] = 0.0;           /* set self distance to zero */
      for (k = 0; k < i; k++) { /* traverse the cluster pairs */
        d = vec_dist(vec, mat_row(ctrs, k), m);
        p[i*c+k] = p[k*c+i] = d;
        if (d < dsts[i]) dsts[i] = d;
        if (d < dsts[k]) dsts[k] = d;
      }                         /* compute distance between clusters, */
    }                           /* update closest cluster distance */
  }
  for (i = 0; i < c; i++)       /* halfen minimum distances */
    dsts[i] *= 0.5;             /* to closest other cluster */
  return dsts;                  /* return pairwise distances */
}  /* c2cdist() */

/*--------------------------------------------------------------------*/

static double update (MATRIX *ctrs, MATRIX *cagg,
                      MATRIX *data, DIMID *cids)
{                               /* --- compute new cluster centers */
  DIMID  n, i;                  /* number of data points, point index */
  DIMID  c, k;                  /* number of clusters, cluster index */
  DIMID  m;                     /* number of dimensions/vector elems. */
  double d, dmax;               /* (maximum) distance a cluster moved */
  double *vec;                  /* current data point/vector */
  double wgt;                   /* weight of current data point */

  mat_init(cagg, MAT_ZERO, NULL);
  mat_rallwgt(cagg, 0.0);       /* clear aggregation incl. weights */
  n = mat_rowcnt(data);         /* get the number of data points */
  c = mat_rowcnt(ctrs);         /* and the number of clusters */
  m = mat_colcnt(ctrs);         /* and the number of dimensions */
  for (i = 0; i < n; i++) {     /* traverse the data points */
    vec = mat_row(data, i);     /* get the data point */
    wgt = mat_rgetwgt(data, i); /* and its weight */
    mat_rowaddv(cagg, cids[i], wgt, vec);
    mat_raddwgt(cagg, cids[i], wgt);
  }                             /* aggregate data point and weight */
  for (dmax = 0, k = 0; k < c; k++) {
    wgt = mat_rgetwgt(cagg, k); /* traverse the clusters */
    if (wgt > 0) mat_rowmuls(cagg, k, 1.0/wgt);
    else         mat_rowset(cagg, k, mat_row(ctrs, k));
    d = vec_dist(mat_row(cagg, k), mat_row(ctrs, k), m);
    mat_rsetwgt(cagg, k, d);    /* compute vectors means (centers) */
    if (d > dmax) dmax = d;     /* and how far the centers moved */
  }                             /* from their previous location */
  return dmax;                  /* return maximum center movement */
}  /* update() */

/*--------------------------------------------------------------------*/

DIMID cls_xhcm (MATRIX *ctrs, MATRIX *data, int method, DIMID epochs,
                DIMID *outids, DIMID outcnt, double size, DIMID *cids)
{                               /* --- (bounds-b.) exact hard c-means */
  int    opt;                   /* parameter update/optim. method */
  int    mod;                   /* parameter update/optim. modifier */
  DIMID  e;                     /* loop variable for epochs */
  DIMID  n, i;                  /* number of data points, point index */
  DIMID  c, k, j, o, p, q;      /* number of clusters, cluster index */
  DIMID  m;                     /* number of dimensions/vector elems. */
  DIMID  z = 1;                 /* size of heap or number of groups */
  DIMID  *cx, *cb, *ci, *ch;    /* indices    of (closest) clusters */
  double *dx, *db;              /* distances  to (closest) clusters */
  DIMID  *cc, *ck, *cg;         /* indices    of closest clusters */
  double *dc, *dk, *dl;         /* distances  between    clusters */
  double *dsts;                 /* data point to cluster distances */
  double *dr;                   /* direction for stratum algorithm */
  double d;                     /* distance to cluster center */
  double dmax;                  /* maximum movement of a cluster */
  double r, t;                  /* radius of search ball for centers */
  char   *xc;                   /* flags for excluded clusters */
  MATRIX *cagg, *mat;           /* to aggregate for new centers */
  double *vec = NULL;           /* current data point */
  #ifdef BENCH                  /* if benchmark version */
  size_t distcnt = 0;           /* number of distance computations */
  #endif                        /* (incremented for each vec_dist) */

  assert(ctrs && data           /* check the function arguments */
  &&    (mat_colcnt(ctrs) == mat_colcnt(data))
  &&    (mat_rowcnt(ctrs) > 0) && (mat_rowcnt(data) > 0));
  n = mat_rowcnt(data);         /* get the number of data points */
  c = mat_rowcnt(ctrs);         /* and the number of clusters */
  m = mat_colcnt(ctrs);         /* and the number of dimensions */

  if (epochs <= 0) {            /* if not to update the clusters */
    if (!cids) return 0;        /* no assignments requested -> abort */
    for (i = 0; i < n; i++) {   /* traverse the data points */
      vec = mat_row(data, i);   /* compute distance to first cluster */
      r   = vec_dist(mat_row(ctrs, 0), vec, m);
      for (k = cids[i] = 0; ++k < c; ) { /* traverse the centers */
        d = vec_dist(mat_row(ctrs, k), vec, m);
        if (d < r) { r = d; cids[i] = k; }
      }                         /* find the closest cluster center */
    }                           /* and store the cluster index */
    return 0;                   /* return the number of epochs */
  }                             /* (no updates have been computed) */

  opt = method & CLS_METHOD;    /* get update method */
  mod = method & CLS_MODIFIER;  /* and update modifier */
  if      (opt == CLS_COGRID) { /* if coordinate grid algorithm */
    if (size == 0) size = 1.0;  /* set a default number of dims. */
    z = (size <= 0) ? (DIMID)-size : (DIMID)(size *(double)m);
    if (z > m) z = m;           /* get the number of dimensions */
    if (z < 1) z = 1; }         /* and ensure a valid value */
  else if ((opt == CLS_HYBRID)  /* if hybrid standard algorithm */
  ||       (opt == CLS_HYSHALL) /* or hybrid shallot algorithm */
  ||       (opt == CLS_YINYANG)) {     /* or yinyang algorithm */
    if (size == 0)              /* set a default value */
      size = (opt == CLS_YINYANG) ? 0.1 : -3;
    z = (size <= 0) ? (DIMID)-size : (DIMID)(size *(double)c);
    if (z >= c) opt = CLS_ELKAN;   /* get number of clusters/groups */
    if (z <= 1) opt = CLS_HAMERLY; /* for the extreme values, switch */
  }                             /* to corresponding simpler algorithm */
  cagg = mat_createx(c, m, MAT_ROWWGTS);
  if (!cagg) return -1;         /* create an aggregation matrix */

  /* --- c distance bounds: [Elkan 2003] --- */
  if (opt == CLS_ELKAN) {       /* method from [Elkan 2003] */
    dsts = dx = (double*)malloc((size_t)(n*c+c) *sizeof(double)
                               +(size_t) n      *sizeof(DIMID));
    if (!dsts) { mat_delete(cagg); return -1; }
    dl = dx +n*c;               /* allocate arrays for distances */
    ci = (DIMID*)(dl +c);       /* and indices of closest clusters */
    if (!(mod & CLS_C2CDIST))   /* if no cluster to cluster distances */
      memset(dl, 0, (size_t)c *sizeof(double));
    #ifdef VERBOSE              /* if verbose progress reporting */
    fprintf(stderr, "       0\b\b\b\b\b\b\b\b");
    #endif                      /* (first epoch is special) */
    db = dx;                    /* epoch 0: traverse the data points */
    for (i = 0; i < n; db += c, i++) {
      vec = mat_row(data, i);   /* compute distance to first center */
      db[ci[i] = 0] = vec_dist(mat_row(ctrs, 0), vec, m);
      for (k = 0; ++k < c; ) {  /* traverse remaining cluster centers */
        db[k] = vec_dist(mat_row(ctrs, k), vec, m);
        if (db[k] < db[ci[i]]) ci[i] = k;
      }                         /* compute distance to next center, */
    }                           /* update index of closest cluster */
    COUNTDC(n*c);               /* count distance computations */
    dmax = update(ctrs, cagg, data, ci);
    COUNTDC(c);                 /* update the cluster centers and */
    SWAP(cagg, ctrs, mat);      /* exchange the center matrices */

    for (e = 0; (++e < epochs) && (dmax > 0); ) {
      #ifdef VERBOSE            /* if verbose progress reporting */
      if ((e & 0x00ff) == 0)    /* print the current number of epochs */
        fprintf(stderr, "%8"DIMID_FMT"\b\b\b\b\b\b\b\b", e);
      #endif
      if (mod & CLS_C2CDIST) {  /* if cluster to cluster distances */
        c2cdist(ctrs, dl, 0);   /* for each cluster: */
        COUNTDC(c*(c-1)/2);     /* compute (half) the distance */
      }                         /* to the closest other cluster */
      db = dx;                  /* traverse the data points */
      for (i = 0; i < n; db += c, i++) {
        o = ci[i];              /* update bound for closest cluster */
        db[o] += mat_rgetwgt(ctrs, o);
        for (d = INFINITY, k = 0; k < c; k++) {
          if (k == o) continue; /* traverse non-closest clusters */
          db[k] -= mat_rgetwgt(ctrs, k);
          if (db[k] < d) d = db[k];   /* find bound for distance */
        }                       /* to the second closest cluster */
        if (d < dl[o])          /* if half the distance to closest */
          d = dl[o];            /* other cluster is a better bound */
        if (db[o] <= d)         /* if closest cluster is the same, */
          continue;             /* no distance comps. are needed */
        vec = mat_row(data, i); /* get the data point to process */
        db[o] = vec_dist(mat_row(ctrs, o), vec, m);
        COUNTDC(1);             /* make upper distance bound tight */
        if (db[o] <= d)         /* if closest cluster is the same, */
          continue;             /* no distance comps. are needed */
        for (k = 0; k < c; k++){/* traverse the cluster centers */
          if ((k == o) || (db[ci[i]] <= db[k]))
            continue;           /* check if center can be closest */
          db[k] = vec_dist(mat_row(ctrs, k), vec, m);
          COUNTDC(1);           /* compute distance to next center */
          if (db[k] < db[ci[i]]) ci[i] = k;
        }                       /* if the current center is closer, */
      }                         /* update index of closest cluster */
      dmax = update(ctrs, cagg, data, ci);
      COUNTDC(c);               /* update the cluster centers and */
      SWAP(cagg, ctrs, mat);    /* exchange the center matrices */
    } }                         /* (current centers and aggregation) */

  /* --- 2 distance bounds: [Hamerly 2010] --- */
  else if (opt == CLS_HAMERLY){ /* method from [Hamerly 2010] */
    o = (mod == CLS_C2CALL) ? c*c : 0;
    dsts = dx = (double*)malloc((size_t)(n+n+c+o) *sizeof(double)
                               +(size_t)(n+o)     *sizeof(DIMID));
    if (!dsts) { mat_delete(cagg); return -1; }
    dl = dx +n +n;              /* allocate arrays for distances */
    dc = dl +c;                 /* to closest cluster centers */
    ci = (DIMID*)(dc +o);       /* and for closest cluster indices */
    cc = ci +n;                 /* for data points and clusters */
    if (!(mod & CLS_C2CDIST))   /* if no cluster to cluster distances */
      memset(dl, 0, (size_t)c *sizeof(double));
    if (mod == CLS_C2CALL)      /* initialize sorted cluster lists */
      for (ck = cc, k = 0; k < c; ck += c, k++)
        for (ck[i = j = 0] = k; i < c; i++)
          if (i != k) ck[++j] = i;
    #ifdef VERBOSE              /* if verbose progress reporting */
    fprintf(stderr, "       0\b\b\b\b\b\b\b\b");
    #endif                      /* (first epoch is special) */
    db = dx;                    /* epoch 0: traverse the data points */
    for (i = 0; i < n; db += 2, i++) {
      vec = mat_row(data, i);   /* compute distance to first center */
      db[1] = vec_dist(mat_row(ctrs, ci[i] = 0), vec, m);
      db[0] = INFINITY;         /* init. distance to second closest */
      for (k = 0; ++k < c; ) {  /* traverse the rem. cluster centers */
        d = vec_dist(mat_row(ctrs, k), vec, m);
        if      (d < db[1]) { db[0] = db[1]; db[1] = d; ci[i] = k; }
        else if (d < db[0]) { db[0] = d; }
      }                         /* update index of closest cluster */
    }                           /* and distances to closest clusters */
    COUNTDC(n*c);               /* count distance computations */
    dmax = update(ctrs, cagg, data, ci);
    COUNTDC(c);                 /* update the cluster centers and */
    SWAP(cagg, ctrs, mat);      /* exchange the center matrices */

    for (e = 0; (++e < epochs) && (dmax > 0); ) {
      #ifdef VERBOSE            /* if verbose progress reporting */
      if ((e & 0x00ff) == 0)    /* print current number of epochs */
        fprintf(stderr, "%8"DIMID_FMT"\b\b\b\b\b\b\b\b", e);
      #endif
      if (mod & CLS_C2CDIST) {  /* if cluster to cluster distances */
        c2cdist(ctrs, dl, (mod == CLS_C2CALL) ? 2 : 0);
        COUNTDC(c*(c-1)/2);     /* comp. dists. to closest cluster */
        /* This is needed for exponion [Newling and Fleuret 2016]: */
        if (mod == CLS_C2CALL)  /* if to use all pairwise distances, */
          for (k = 0; k < c; k++)       /* sort clusters by distance */
            dim_isort(cc+k*c+1, (size_t)c-1, +1, dc+k*c);
      }                         /* (set up sorted cluster lists) */
      db = dx;                  /* traverse the data points */
      for (i = 0; i < n; db += 2, i++) {
        o = ci[i];              /* update bound for closest cluster */
        db[1] += mat_rgetwgt(ctrs, o);
        db[0] -= dmax;          /* update bound for other clusters */
        d = (db[0] > dl[o])     /* if half the distance to closest */
          ?  db[0] : dl[o];     /* other cluster is a better bound */
        if (db[1] <= d)         /* if closest cluster is the same, */
          continue;             /* no distance comps. are needed */
        vec = mat_row(data, i); /* get the data point to process */
        db[1] = vec_dist(mat_row(ctrs, o), vec, m);
        COUNTDC(1);             /* make upper distance bound tight */
        if (db[1] <= d)         /* if closest cluster is the same, */
          continue;             /* no distance comps. are needed */
        db[0] = INFINITY;       /* reinit. second closest distance */
        if (mod == CLS_C2CALL){ /* if cluster to cluster distances */
          /* This is the exponion version [Newling and Fleuret 2016]: */
          /* clusters are sorted by their distance to each cluster */
          /* and hence only the first clusters in the sorted list, */
          /* up to the search ball radius r, have to be checked.   */
          dk = dc +o*c;         /* get needed row of distance matrix */
          ck = cc +o*c;         /* and (sorted) cluster index matrix */
          r  = (db[1]+dl[o]) *2 /* get radius of ball around cluster */
             * (1+4*DBL_EPSILON);   /* (account for rounding errors) */
          for (j = 0; (++j < c) && (dk[k = ck[j]] <= r); ) {
            d = vec_dist(mat_row(ctrs, k), vec, m);
            COUNTDC(1);         /* compute distance to next cluster */
            if      (d < db[1]) { db[0] = db[1]; db[1] = d; ci[i] = k; }
            else if (d < db[0]) { db[0] = d; }
          } }                   /* find distances to closest clusters */
        else {                  /* if to recompute all distances */
          /* This is the standard version of [Hamerly 2010]: */
          /* all data point to cluster distances are recomputed. */
          for (k = 0; k < c; k++) { /* traverse the cluster centers, */
            if (k == o) continue;   /* but skip (old) closest cluster */
            d = vec_dist(mat_row(ctrs, k), vec, m);
            if      (d < db[1]) { db[0] = db[1]; db[1] = d; ci[i] = k; }
            else if (d < db[0]) { db[0] = d; }
          }                     /* update index of closest cluster */
          COUNTDC(c-1);         /* and distances to closest clusters */
        }                       /* (i.e. reinitialize the bounds) */
      }  /* for (i = 0; i < n; i++) .. */
      dmax = update(ctrs, cagg, data, ci);
      COUNTDC(c);               /* update the cluster centers and */
      SWAP(cagg, ctrs, mat);    /* exchange the center matrices */
    } }                         /* (current centers and aggregation) */

  /* --- 2 distance bounds: shallot algorithm --- */
  else if (opt == CLS_SHALLOT){ /* modified exponion algorithm */
    dsts = dx = (double*)malloc((size_t)(n+n+c+c*c) *sizeof(double)
                               +(size_t)(n+n+c*c)   *sizeof(DIMID));
    if (!dsts) { mat_delete(cagg); return -1; }
    dl = dx +n +n;              /* allocate arrays for distances */
    dc = dl +c;                 /* to closest cluster centers */
    ci = (DIMID*)(dc +c*c);     /* and for closest cluster indices */
    ch = ci +n;                 /* for data points and clusters and */
    cc = ch +n;                 /* for sorted cluster lists */
    for (ck = cc, k = 0; k < c; ck += c, k++)
      for (ck[i = j = 0] = k; i < c; i++)
        if (i != k) ck[++j] = i;/* initialize sorted cluster lists */
    #ifdef VERBOSE              /* if verbose progress reporting */
    fprintf(stderr, "       0\b\b\b\b\b\b\b\b");
    #endif                      /* (first epoch is special) */
    db = dx;                    /* epoch 0: traverse the data points */
    for (i = 0; i < n; db += 2, i++) {
      vec = mat_row(data, i);   /* compute distance to first center */
      db[1] = vec_dist(mat_row(ctrs, ch[i] = ci[i] = 0), vec, m);
      db[0] = INFINITY;         /* init. distance to second closest */
      for (k = 0; ++k < c; ) {  /* traverse the rem. cluster centers */
        d = vec_dist(mat_row(ctrs, k), vec, m);
        if      (d < db[1]) { db[0] = db[1]; ch[i] = ci[i];
                              db[1] = d;     ci[i] = k; }
        else if (d < db[0]) { db[0] = d;     ch[i] = k; }
      }                         /* update index of closest clusters */
    }                           /* and distance to closest clusters */
    COUNTDC(n*c);               /* count distance computations */
    dmax = update(ctrs, cagg, data, ci);
    COUNTDC(c);                 /* update the cluster centers and */
    SWAP(cagg, ctrs, mat);      /* exchange the center matrices */

    for (e = 0; (++e < epochs) && (dmax > 0); ) {
      #ifdef VERBOSE            /* if verbose progress reporting */
      if ((e & 0x00ff) == 0)    /* print current number of epochs */
        fprintf(stderr, "%8"DIMID_FMT"\b\b\b\b\b\b\b\b", e);
      #endif
      c2cdist(ctrs, dl, 2);     /* for each cluster: */
      COUNTDC(c*(c-1)/2);       /* compute (half) the distance */
      for (k = 0; k < c; k++)   /* to the closest other cluster */
        dim_isort(cc+k*c+1, (size_t)c-1, +1, dc+k*c);
      db = dx;                  /* traverse the data points */
      for (i = 0; i < n; db += 2, i++) {
        o = ci[i];              /* update bound for closest cluster */
        db[1] += mat_rgetwgt(ctrs, o);
        db[0] -= dmax;          /* update bound for other clusters */
        d = (db[0] > dl[o])     /* if half the distance to closest */
          ?  db[0] : dl[o];     /* other cluster is a better bound */
        if (db[1] <= d)         /* if closest cluster is the same, */
          continue;             /* no distance comps. are needed */
        vec = mat_row(data, i); /* get the data point to process */
        db[1] = t = vec_dist(mat_row(ctrs, o), vec, m);
        COUNTDC(1);             /* make upper distance bound tight */
        if (t <= d)             /* if closest cluster is the same, */
          continue;             /* no distance comps. are needed */
        p = ch[i];              /* get (old) 2nd closest index */
        db[0] = vec_dist(mat_row(ctrs, p), vec, m);
        COUNTDC(1);             /* make second distance bound tight */
        if (db[0] < t) {        /* if 2nd cluster has moved closer */
          t = db[0]; db[0] = db[1]; db[1] = t;
          ci[i] = p; ch[i] = p = o; o = ci[i];
        }                       /* exchange the two clusters */
        dk = dc +o*c;           /* get needed row of distance matrix */
        ck = cc +o*c;           /* and (sorted) cluster index matrix */
        r  = (t+dl[o]) *2       /* get radius of ball around cluster */
           * (1+4*DBL_EPSILON); /* (account for rounding errors) */
        d  = (t+db[0])          /* get alternative search ball radius */
           * (1+4*DBL_EPSILON); /* (account for rounding errors) */
        if (d < r) r = d;       /* choose smaller search ball radius */
        for (j = 0; (++j < c) && (dk[k = ck[j]] <= r); ) {
          if (k == p) continue; /* skip old 2nd closest cluster */
          d = vec_dist(mat_row(ctrs, k), vec, m);
          COUNTDC(1);           /* compute distance to next cluster */
          if      (d < db[1]) { /* if closer than old closest */
            db[0] = db[1]; ch[i] = ci[i]; db[1] = d; ci[i] = k;
            if ((d = (t+db[0]) *(1+4*DBL_EPSILON)) < r) r = d; }
          else if (d < db[0]) { /* if closer than old 2nd closest */
            db[0] = d;     ch[i] = k;
            if ((d = (t+d)     *(1+4*DBL_EPSILON)) < r) r = d; }
        }                       /* find distances to closest clusters */
      }                         /* and index of closest cluster */
      dmax = update(ctrs, cagg, data, ci);
      COUNTDC(c);               /* update the cluster centers and */
      SWAP(cagg, ctrs, mat);    /* exchange the center matrices */
    } }                         /* (current centers and aggregation) */

  /* --- 2 distance bounds: annular algorithm [Drake 2013] --- */
  else if ((opt == CLS_ANNULAR) /* method from [Drake 2013] or */
  ||       (opt == CLS_STRATUM)) {     /* stratum modification */
    o = (opt == CLS_STRATUM) ? m : 0;
    dsts = dx = (double*)malloc((size_t)(n+n+n+c+c+o) *sizeof(double)
                               +(size_t)(n+n+c)       *sizeof(DIMID));
    if (!dsts) { mat_delete(cagg); return -1; }
    dk = dx +n +n;              /* allocate arrays for distances */
    dl = dk +n;                 /* to closest cluster centers, */
    dc = dl +c;                 /* and for data point norms/projs., */
    dr = dc +c;                 /* for the stratum direction, and */
    ci = (DIMID*)(dr +o);       /* for closest cluster indices */
    ch = ci +n;                 /* for data points and clusters and */
    cg = ch +n;                 /* for norm/proj. sorted clusters */
    if (!(mod & CLS_C2CDIST))   /* if no cluster to cluster distances */
      memset(dl, 0, (size_t)c *sizeof(double));
    for (k = 0; k < c; k++)     /* init. the cluster list to be */
      cg[k] = k;                /* sorted by cluster center norms */

    if (opt == CLS_STRATUM) {   /* if stratum variant (vs. annulus) */
      /* The idea of the stratum algorithm is similar to the annular  */
      /* algorithm, only that the norms (distances to the origin) are */
      /* replaced by the projections to a chosen direction, and thus  */
      /* strata are used instead of annuli to restrict the centers to */
      /* which distances need to be computed to find the two closest. */
      /* The direction may be chosen in different ways, for example,  */
      /* by finding the first principal component of the centers.     */
      /* For efficiency, here two extreme centers are determined and  */
      /* the direction vector from one center to the other is chosen. */
      vec = mat_row(ctrs, 0);   /* get a (random) cluster center */
      for (r = 0, o = k = 0; ++k < c; ) {
        d = vec_dist(mat_row(ctrs, k), vec, m);
        if (d > r) { r = d; o = k; }
      }                         /* traverse other cluster centers, */
      COUNTDC(c-1);             /* find and copy farthest center */
      vec_copy(dr, mat_row(ctrs, o), m);
      for (r = 0, o = k = 0; ++k < c; ) {
        if (k == o) continue;   /* traverse other cluster centers */
        d = vec_dist(mat_row(ctrs, k), dr, m);
        if (d > r) { r = d; o = k; }
      }                         /* find cluster center farthest */
      COUNTDC(c-1);             /* from the one found in 1st step */
      vec_add(dr, m, dr, -1, mat_row(ctrs, o));
      vec_unitlen(dr, dr, m);   /* form and normalize */
    }                           /* the center difference vector */

    #ifdef VERBOSE              /* if verbose progress reporting */
    fprintf(stderr, "       0\b\b\b\b\b\b\b\b");
    #endif                      /* (first epoch is special) */
    db = dx;                    /* epoch 0: traverse the data points */
    for (i = 0; i < n; db += 2, i++) {
      vec = mat_row(data, i);   /* compute distance to first center */
      db[1] = vec_dist(mat_row(ctrs, ch[i] = ci[i] = 0), vec, m);
      db[0] = INFINITY;         /* init. distance to second closest */
      dk[i] = (opt == CLS_ANNULAR)  /* and norm/projection of point */
            ? vec_len(vec, m) : vec_mul(vec, dr, m);
      for (k = 0; ++k < c; ) {  /* traverse the rem. cluster centers */
        d = vec_dist(mat_row(ctrs, k), vec, m);
        if      (d < db[1]) { db[0] = db[1]; ch[i] = ci[i];
                              db[1] = d;     ci[i] = k; }
        else if (d < db[0]) { db[0] = d;     ch[i] = k; }
      }                         /* update index of closest clusters */
    }                           /* and distance to closest clusters */
    COUNTDC(n*c);               /* count distance computations */
    dmax = update(ctrs, cagg, data, ci);
    COUNTDC(c);                 /* update the cluster centers and */
    SWAP(cagg, ctrs, mat);      /* exchange the center matrices */

    for (e = 0; (++e < epochs) && (dmax > 0); ) {
      #ifdef VERBOSE            /* if verbose progress reporting */
      if ((e & 0x00ff) == 0)    /* print current number of epochs */
        fprintf(stderr, "%8"DIMID_FMT"\b\b\b\b\b\b\b\b", e);
      #endif
      if (mod & CLS_C2CDIST) {  /* if cluster to cluster distances */
        c2cdist(ctrs, dl, 0);   /* for each cluster: */
        COUNTDC(c*(c-1)/2);     /* compute (half) the distance */
      }                         /* to the closest other cluster */
      if (opt == CLS_ANNULAR) { /* if annular algorithm [Drake 2013] */
        for (k = 0; k < c; k++) /* compute cluster center norms */
          dc[k] = vec_len(mat_row(ctrs, k), m); }
      else {                    /* if stratum algorithm */
        for (k = 0; k < c; k++) /* compute center projection lengths */
          dc[k] = vec_mul(mat_row(ctrs, k), dr, m);
      }                         /* (cluster center reference values) */
      COUNTDC(c);               /* count the norm/projection comps. */
      i2d_qsort(cg, (size_t)c, +1, dc); /* sort cluster centers */
      db = dx;                  /* traverse the data points */
      for (i = 0; i < n; db += 2, i++) {
        o = ci[i];              /* update bound for closest cluster */
        db[1] += mat_rgetwgt(ctrs, o);
        db[0] -= dmax;          /* update bound for other clusters */
        d = (db[0] > dl[o])     /* if half the distance to closest */
          ?  db[0] : dl[o];     /* other cluster is a better bound */
        if (db[1] <= d)         /* if closest cluster is the same, */
          continue;             /* no distance comps. are needed */
        vec = mat_row(data, i); /* get the data point to process */
        db[1] = vec_dist(mat_row(ctrs, o), vec, m);
        COUNTDC(1);             /* make upper distance bound tight */
        if (db[1] <= d)         /* if closest cluster is the same, */
          continue;             /* no distance comps. are needed */
        p = ch[i];              /* note index of old 2nd closest */
        db[0] = vec_dist(mat_row(ctrs, p), vec, m);
        COUNTDC(1);             /* get dist. to 2nd closest cluster */
        if (db[1] > db[0]) {    /* if old 2nd closest is now closer */
          d = db[1]; db[1] = db[0]; db[0] = d;
          ci[i] = p; ch[i] = o; /* exchange the closest and the */
        }                       /* second closest cluster */
        /* All centers with a norm between r-d and r+d have to be   */
        /* checked, where r = ||x_i|| and d = max(db[0], db[1]), or */
        /* all centers in the stratum between r-d and r+d have to   */
        /* be checked, where r = x_i *dr and d = max(db[0], db[1]). */
        r = dk[i];              /* get norm/projection of data point */
        j = dim_ibisect(r +db[0], cg, (size_t)c, dc);
        r -= db[0];             /* get norm bound from 2nd closest */
        while ((--j >= 0) && (dc[k = cg[j]] >= r)) {
          if ((k == o) || (k == p)) /* traverse clusters in range, */
            continue;           /* but skip (old) closest clusters */
          d = vec_dist(mat_row(ctrs, k), vec, m);
          COUNTDC(1);           /* compute distance to next cluster */
          if      (d < db[1]) { db[0] = db[1]; ch[i] = ci[i];
                                db[1] = d;     ci[i] = k; }
          else if (d < db[0]) { db[0] = d;     ch[i] = k; }
        }                       /* find distances to closest clusters */
      }                         /* and index of closest cluster */
      dmax = update(ctrs, cagg, data, ci);
      COUNTDC(c);               /* update the cluster centers and */
      SWAP(cagg, ctrs, mat);    /* exchange the center matrices */
    } }                         /* (current centers and aggregation) */

  /* --- 2 distance bounds: coordinate grid algorithm --- */
  else if (opt == CLS_COGRID) { /* coordinate grid method */
    dsts = dx = (double*)malloc((size_t)(n+n+c)       *sizeof(double)
                               +(size_t)(n+n+c+c+z*c) *sizeof(DIMID));
    if (!dsts) { mat_delete(cagg); return -1; }
    dl = dx +n +n;              /* allocate arrays for distances */
    ci = (DIMID*)(dl +c);       /* to closest cluster centers, */
    ch = ci +n;                 /* and for closest cluster indices */
    cc = ch +n;                 /* and for collected clusters */
    cg = cc +c;                 /* and their collected flags */
    cx = cg +c;                 /* and dimension sorted centers */
    if (!(mod & CLS_C2CDIST))   /* if no cluster to cluster distances */
      memset(dl, 0, (size_t)c *sizeof(double));
    memset(cg, 0, (size_t)c *sizeof(DIMID));

    #ifdef VERBOSE              /* if verbose progress reporting */
    fprintf(stderr, "       0\b\b\b\b\b\b\b\b");
    #endif                      /* (first epoch is special) */
    db = dx;                    /* epoch 0: traverse the data points */
     for (i = 0; i < n; db += 2, i++) {
      vec = mat_row(data, i);   /* compute distance to first center */
      db[1] = vec_dist(mat_row(ctrs, ch[i] = ci[i] = 0), vec, m);
      db[0] = INFINITY;         /* init. distance to second closest */
      for (k = 0; ++k < c; ) {  /* traverse the rem. cluster centers */
        d = vec_dist(mat_row(ctrs, k), vec, m);
        if      (d < db[1]) { db[0] = db[1]; ch[i] = ci[i];
                              db[1] = d;     ci[i] = k; }
        else if (d < db[0]) { db[0] = d;     ch[i] = k; }
      }                         /* update index of closest clusters */
    }                           /* and distance to closest clusters */
    COUNTDC(n*c);               /* count distance computations */
    dmax = update(ctrs, cagg, data, ci);
    COUNTDC(c);                 /* update the cluster centers and */
    SWAP(cagg, ctrs, mat);      /* exchange the center matrices */

    for (e = 0; (++e < epochs) && (dmax > 0); ) {
      #ifdef VERBOSE            /* if verbose progress reporting */
      if ((e & 0x00ff) == 0)    /* print current number of epochs */
        fprintf(stderr, "%8"DIMID_FMT"\b\b\b\b\b\b\b\b", e);
      #endif
      if (mod & CLS_C2CDIST) {  /* if cluster to cluster distances */
        c2cdist(ctrs, dl, 0);   /* for each cluster: */
        COUNTDC(c*(c-1)/2);     /* compute (half) the distance */
      }                         /* to the closest other cluster */
      ck = cx;                  /* sort centers by input dimension */
      for (k = 0; k < z; ck += c, k++)
        dim_psort(ck, (size_t)c, +1, mat_rows(ctrs),
                  dimcmp, (void*)(diff_t)k);
      db = dx;                  /* traverse the data points */
      for (i = 0; i < n; db += 2, i++) {
        o = ci[i];              /* get (old) closest cluster index */
        db[1] += mat_rgetwgt(ctrs, o);
        db[0] -= dmax;          /* update cluster distance bounds */
        d = (db[0] > dl[o])     /* if half the distance to closest */
          ?  db[0] : dl[o];     /* other cluster is a better bound */
        if (db[1] <= d)         /* if closest cluster is the same, */
          continue;             /* no distance comps. are needed */
        vec = mat_row(data, i); /* get the data point to process */
        db[1] = vec_dist(mat_row(ctrs, o), vec, m);
        COUNTDC(1);             /* make upper distance bound tight */
        if (db[1] <= d)         /* if closest cluster is the same, */
          continue;             /* no distance comps. are needed */
        p = ch[i];              /* note index of old 2nd closest */
        db[0] = vec_dist(mat_row(ctrs, p), vec, m);
        COUNTDC(1);             /* get dist. to 2nd closest cluster */
        if (db[1] > db[0]) {    /* if old 2nd closest is now closer */
          d = db[1]; db[1] = db[0]; db[0] = d;
          ci[i] = p; ch[i] = o; /* exchange the (old) closest and */
        }                       /* the (old) second closest cluster */
        t = vec[0] +db[0];      /* get norm bound from 2nd closest */
        r = vec[0] -db[0];      /* and compute the value range */
        j = bisect(t, ck = cx, c, mat_rows(ctrs), 0);
        for (q = 0; (--j >= 0) && (mat_get(ctrs, ck[j], 0) >= r); q++)
          cg[cc[q] = ck[j]] = 1;/* collect qualifying clusters */
        for (k = 0; ++k < z; ){ /* traverse the remaining dimensions */
          t = vec[k] +db[0];    /* get norm bound from 2nd closest */
          r = vec[k] -db[0];    /* and compute the value range */
          j = bisect(t, ck += c, c, mat_rows(ctrs), k);
          while ((--j >= 0) && (mat_get(ctrs, ck[j], k) >= r))
            if (cg[ck[j]]) cg[ck[j]]++;
        }                       /* filter collected cluster centers */
        for (k = j = 0; k < q; k++) {
          if (cg[cc[k]] >= z)   /* traverse collected centers and */
            cc[j++] = cc[k];    /* remove filtered out centers */
          cg[cc[k]] = 0;        /* clear all set cluster flags */
        }                       /* (that were set for first dim.) */
        cg[o] = cg[p] = 0;      /* clear flags for old closest */
        for (q = 0; q < j; q++) {
          k = cc[q];            /* traverse the collected centers */
          if ((k == o) || (k == p))
            continue;           /* skip (old) closest clusters */
          d = vec_dist(mat_row(ctrs, k), vec, m);
          COUNTDC(1);           /* compute distance to next cluster */
          if      (d < db[1]) { db[0] = db[1]; ch[i] = ci[i];
                                db[1] = d;     ci[i] = k; }
          else if (d < db[0]) { db[0] = d;     ch[i] = k; }
        }                       /* find distances to closest clusters */
      }                         /* and index of closest cluster */
      dmax = update(ctrs, cagg, data, ci);
      COUNTDC(c);               /* update the cluster centers and */
      SWAP(cagg, ctrs, mat);    /* exchange the center matrices */
    } }                         /* (current centers and aggregation) */

  /* --- z+1 distance bounds: hybrid approach --- */
  else if ((opt == CLS_HYBRID)  /* if hybrid of Elkan and Hamerly */
  ||       (opt == CLS_HYSHALL)) { /* w/ or w/o shallot algorithm */
    if (opt == CLS_HYSHALL) mod = CLS_C2CALL;
    o = (mod == CLS_C2CALL) ? c*c : 0;
    dsts = dx = (double*)malloc((size_t)(n*(z+1)+c+o) *sizeof(double)
                               +(size_t)(n*(z+1)+o)   *sizeof(DIMID)
                               +(size_t) c            *sizeof(char));
    if (!dsts) { mat_delete(cagg); return -1; }
    dl = dx +n*(z+1);           /* allocate arrays for distances */
    dc = dl +c;                 /* to closest cluster centers */
    ci = (DIMID*)(dc +o);       /* and for closest cluster indices */
    cx = ci +n;                 /* for data points and clusters */
    cc = cx +n*z;               /* (elements of cluster heaps) and */
    xc = (char*)(cc +o);        /* distance sorted cluster indices */
    memset(cx, 0, (size_t)(n*z) *sizeof(DIMID));
    if (!(mod & CLS_C2CDIST))   /* if no cluster to cluster distances */
      memset(dl, 0, (size_t)c *sizeof(double));
    if (mod == CLS_C2CALL)      /* initialize sorted cluster lists */
      for (ck = cc, k = 0; k < c; ck += c, k++)
        for (ck[i = j = 0] = k; i < c; i++)
          if (i != k) ck[++j] = i;

    #ifdef VERBOSE              /* if verbose progress reporting */
    fprintf(stderr, "       0\b\b\b\b\b\b\b\b");
    #endif                      /* (first epoch is special) */
    db = dx; cb = cx;           /* epoch 0: traverse the data points */
    #if SORTED==1
    if (z <= THRESH) {          /* if only few closest clusters */
      for (i = 0; i < n; db += z+1, cb += z, i++) {
        vec = mat_row(data, i); /* compute distance to first center */
        db[z] = vec_dist(mat_row(ctrs, ci[i] = 0), vec, m);
        db[0] = INFINITY;       /* init. distance to second closest */
        j = z;                  /* init. heap/sorted list variables */
        for (k = 0; ++k < c; ){ /* traverse the rem. cluster centers */
          d = vec_dist(mat_row(ctrs, k), vec, m);
          if (j > 0) j -= 1;    /* compute distance to next cluster */
          if (d < db[z]) {      /* if closer than old closest */
            for (p = j; ++p < z; ) {
              db[p-1] = db[p]; cb[p-1] = cb[p]; }
            db[z-1] = db[z]; db[z] = d;
            cb[z-1] = ci[i]; ci[i] = k; }
          else if (d < db[0]) { /* if only some closest cluster */
            for (p = j; d < db[++p]; ) {
              db[p-1] = db[p]; cb[p-1] = cb[p]; }
            db[p-1] = d; cb[p-1] = k;
          }                     /* update distances of old closest */
        }                       /* and bring them into desc. order */
      } }  /* if (z <= THRESH) .. */
    else {                      /* if many closest centers */
    #endif
      for (i = 0; i < n; db += z+1, cb += z, i++) {
        vec = mat_row(data, i); /* compute distance to first center */
        db[z] = vec_dist(mat_row(ctrs, ci[i] = 0), vec, m);
        db[0] = INFINITY;       /* init. distance to second closest */
        j = z;                  /* init. heap/sorted list variables */
        for (k = 0; ++k < c; ){ /* traverse the rem. cluster centers */
          d = vec_dist(mat_row(ctrs, k), vec, m);
          if (j > 0) j -= 1;    /* compute distance to next cluster */
          if (d < db[z]) {      /* if closer than closest cluster */
            xsift(db, cb, j, z, db[z], ci[i]);
            db[z] = d; ci[i] = k; }
          else if (d < db[0]) { /* if closer than some other cluster */
            xsift(db, cb, j, z, d, k);
          }                     /* let the second closest distance */
        }                       /* and corresponding cluster index */
      }                         /* sift down in (maximum) heap */
    #if SORTED==1
    }  /* if (z <= THRESH) .. else .. */
    #endif
    COUNTDC(n*c);               /* count distance computations */
    dmax = update(ctrs, cagg, data, ci);
    COUNTDC(c);                 /* update the cluster centers and */
    SWAP(cagg, ctrs, mat);      /* exchange the center matrices */

    for (e = 0; (++e < epochs) && (dmax > 0); ) {
      #ifdef VERBOSE            /* if verbose progress reporting */
      if ((e & 0x00ff) == 0)    /* print the current number of epochs */
        fprintf(stderr, "%8"DIMID_FMT"\b\b\b\b\b\b\b\b", e);
      #endif
      if (mod & CLS_C2CDIST) {  /* if cluster to cluster distances */
        c2cdist(ctrs, dl, (mod == CLS_C2CALL) ? 2 : 0);
        COUNTDC(c*(c-1)/2);     /* comp. dists. to closest cluster */
        /* This is needed for the generalized exponion method: */
        if (mod == CLS_C2CALL)  /* if to use all pairwise distances, */
          for (k = 0; k < c; k++)       /* sort clusters by distance */
            dim_isort(cc+k*c+1, (size_t)c, +1, dc+k*c);
      }                         /* (set up sorted cluster lists) */
      db = dx; cb = cx;         /* traverse the data points */
      for (i = 0; i < n; db += z+1, cb += z, i++) {
        o = ci[i];              /* update bound for closest cluster */
        db[z] += mat_rgetwgt(ctrs, o);
        db[0] -= dmax;          /* update bound for other clusters */
        for (d = db[0], k = z; --k > 0; ) {
          db[k] -= mat_rgetwgt(ctrs, cb[k]);
          if (db[k] < d) d = db[k];   /* find bound for distance */
        }                       /* to the second closest cluster */
        if (d < dl[o])          /* if half the distance to closest */
          d = dl[o];            /* other cluster is a better bound */
        if (db[z] <= d)         /* if closest cluster is the same, */
          continue;             /* no distance comps. are needed */
        vec = mat_row(data, i); /* get the data point to process */
        db[z] = vec_dist(mat_row(ctrs, o), vec, m);
        COUNTDC(1);             /* make upper distance bound tight */
        if (db[z] <= d)         /* if closest cluster is the same, */
          continue;             /* no distance comps. are needed */
        if (db[z] <= db[0]) {   /* if distance bound check succeeds */
          #if 1
          /* The following simpler version works, even though it  */
          /* destroys the distance order/heap structure, because  */
          /* this structure is not needed as long as the distance */
          /* bound check succeeds, but is rebuilt once it fails.  */
          for (k = z; --k > 0;){/* traverse other (close) clusters */
            if (db[z] <= db[k]) continue;
            db[k] = vec_dist(mat_row(ctrs, cb[k]), vec, m);
            COUNTDC(1);         /* make lower distance bound tight */
            if (db[z] <= db[k]) continue;
            d = db[k]; db[k] = db[z]; db[z] = d;
            o = cb[k]; cb[k] = ci[i]; ci[i] = o;
          }                     /* update the closest cluster */
          continue;             /* (note distance and cluster index) */
          #else
          if (z <= THRESH) {    /* if only few closest clusters */
            for (j = z; --j > 0; ) { /* traverse closest clusters */
              d = vec_dist(mat_row(ctrs, o = cb[j]), vec, m);
              if (d < db[z]) {  /* if closer than old closest cluster */
                for (p = j; ++p < z; ) {
                  db[p-1] = db[p]; cb[p-1] = cb[p]; }
                db[z-1] = db[z]; db[z] = d;
                cb[z-1] = ci[i]; ci[i] = o; }
              else {            /* if closer than some other cluster */
                for (p = j; d < db[++p]; ) {
                  db[p-1] = db[p]; cb[p-1] = cb[p]; }
                db[p-1] = d; cb[p-1] = o;
              }                 /* update distances of old closest */
            } }                 /* and bring them into desc. order */
          else {                /* if many closest centers */
            for (j = z; --j >= z >> 1; ) {
              xc[k = cb[j]] = 1;/* rebuild 2nd half of cluster heap */
              db[j] = d = vec_dist(mat_row(ctrs, k), vec, m);
              if (d >= db[z]) continue; /* simply recompute distances */
              db[j] = db[z]; db[z] = d; /* of clusters in the heap; */
              cb[j] = ci[i]; ci[i] = k; /* update closest if needed */
            }                   /* (find closest cluster center) */
            for (j = z >> 1; --j > 0; ) {
              xc[k = cb[j]] = 1;/* rebuild 1st half of cluster heap */
              d = vec_dist(mat_row(ctrs, k), vec, m);
              if (d < db[z]) {  /* if closer than old closest cluster */
                xsift(db, cb, j, z, db[z], ci[i]);
                db[z] = d; ci[i] = k; }
              else {            /* if closer than some other cluster */
                xsift(db, cb, j, z, d, k);
              }                 /* let second closest distance and */
            }                   /* cluster index sift down in heap */
          }
          COUNTDC(z-1);         /* count distance computations */
          continue;             /* continue with next data point */
          #endif
        }
        /* distance bound check failed even with tight bound: */
        if (opt == CLS_HYSHALL){/* if combined with shallot algo. */
          /* This is the generalized shallot version:              */
          /* clusters are sorted by their distance to each cluster */
          /* and hence only the first clusters in the sorted list, */
          /* up to the search ball radius r, have to be checked.   */
          /* The search radius is determined as the minimum of the */
          /* radius for the exponion algorithm and the range of    */
          /* the interval for the annular or stratum algorithms.   */
          memset(xc, 0, (size_t)c *sizeof(char));
          xc[o] = 1;            /* initialize the cluster flags */
          #if SORTED==1
          if (z <= THRESH) {    /* if only few closest clusters */
            for (j = z; --j >= 0; ) { /* traverse closest clusters */
              d = vec_dist(mat_row(ctrs, k = cb[j]), vec, m);
              if (d < db[z]) {  /* if closer than old closest cluster */
                for (p = j; ++p < z; ) {
                  db[p-1] = db[p]; cb[p-1] = cb[p]; }
                db[z-1] = db[z]; db[z] = d;
                cb[z-1] = ci[i]; ci[i] = k; }
              else {            /* if closer than some other cluster */
                for (p = j; d < db[++p]; ) {
                  db[p-1] = db[p]; cb[p-1] = cb[p]; }
                db[p-1] = d; cb[p-1] = k;
              }                 /* update distances of old closest */
            }                   /* and bring them into desc. order */
            COUNTDC(z);         /* count the distance computations */
            o  = ci[i];         /* get (possibly new) closest index */
            dk = dc +o*c;       /* get needed row of distance matrix */
            ck = cc +o*c;       /* and (sorted) cluster index matrix */
            t  = db[z];         /* get dist. to closest as b. radius */
            r  = (t*2+dk[ck[z]])/* get radius of ball around cluster */
               * (1+4*DBL_EPSILON); /* (account for rounding errors) */
            d  = (t  +db[0])    /* get alternative search ball radius */
               * (1+4*DBL_EPSILON); /* (account for rounding errors) */
            if (d < r) r = d;   /* choose smaller search ball radius */
            for (j = 0; (++j < c) && (dk[k = ck[j]] <= r); ) {
              if (xc[k]) continue; /* skip already processed centers */
              d = vec_dist(mat_row(ctrs, k), vec, m);
              COUNTDC(1);       /* compute distance to next cluster */
              if (d < db[z]) {  /* if closer than closest cluster */
                for (p = 0; ++p < z; ) {
                  db[p-1] = db[p]; cb[p-1] = cb[p]; }
                db[z-1] = db[z]; db[z] = d;
                cb[z-1] = ci[i]; ci[i] = k;
                if ((d = (t+db[0]) *(1+4*DBL_EPSILON)) < r) r = d; }
              else if (d < db[0]) { /* if closer than some other */
                for (p = 0; d < db[++p]; ) {
                  db[p-1] = db[p]; cb[p-1] = cb[p]; }
                db[p-1] = d; cb[p-1] = k;
                if ((d = (t+db[0]) *(1+4*DBL_EPSILON)) < r) r = d;
              }                 /* let second closest distance and */
            } }                 /* cluster index sift down in list */
          else {                /* if many closest centers */
          #endif
            for (j = z; --j >= z >> 1; ) {
              xc[k = cb[j]] = 1;/* rebuild 2nd half of cluster heap */
              db[j] = d = vec_dist(mat_row(ctrs, k), vec, m);
              if (d >= db[z]) continue; /* simply recompute distances */
              db[j] = db[z]; db[z] = d; /* of clusters in the heap; */
              cb[j] = ci[i]; ci[i] = k; /* update closest if needed */
            }                   /* (find closest cluster center) */
            for (j = z >> 1; --j >= 0; ) {
              xc[k = cb[j]] = 1;/* rebuild 1st half of cluster heap */
              d = vec_dist(mat_row(ctrs, k), vec, m);
              if (d < db[z]) {  /* if closer than closest cluster */
                xsift(db, cb, j, z, db[z], ci[i]);
                db[z] = d; ci[i] = k; }
              else {            /* if closer than some other cluster */
                xsift(db, cb, j, z, d, k);
              }                 /* let second closest distance and */
            }                   /* cluster index sift down in heap */
            COUNTDC(z);         /* count the distance computations */
            o  = ci[i];         /* get (possibly new) closest index */
            dk = dc +o*c;       /* get needed row of distance matrix */
            ck = cc +o*c;       /* and (sorted) cluster index matrix */
            t  = db[z];         /* get dist. to closest as b. radius */
            r  = (t*2+dk[ck[z]])/* get radius of ball around cluster */
               * (1+4*DBL_EPSILON); /* (account for rounding errors) */
            d  = (t  +db[0])    /* get alternative ball radius */
               * (1+4*DBL_EPSILON); /* (account for rounding errors) */
            if (d < r) r = d;   /* choose smaller search ball radius */
            for (j = 0; (++j < c) && (dk[k = ck[j]] <= r); ) {
              if (xc[k]) continue;/* skip already processed centers */
              d = vec_dist(mat_row(ctrs, k), vec, m);
              COUNTDC(1);       /* compute distance to next cluster */
              if (d < db[z]) {  /* if closer than closest cluster */
                xsift(db, cb, 0, z, db[z], ci[i]);
                db[z] = d; ci[i] = k;
                if ((d = (t+db[0]) *(1+4*DBL_EPSILON)) < r) r = d; }
              else if (d < db[0]) { /* if closer than some other */
                xsift(db, cb, 0, z, d, k);
                if ((d = (t+db[0]) *(1+4*DBL_EPSILON)) < r) r = d;
              }                 /* let second closest distance and */
            }                   /* cluster index sift down in heap */
          #if SORTED==1
          }  /* if (z <= THRESH) .. else .. */
          #endif
        }  /* if (opt == CLS_HYSHALL) .. */
        else if (mod == CLS_C2CALL) { /* if cls. to cls. distances */
          /* This is the generalized exponion version:             */
          /* clusters are sorted by their distance to each cluster */
          /* and hence only the first clusters in the sorted list, */
          /* up to the search ball radius r, have to be checked.   */
          dk = dc +o*c;         /* get needed row of distance matrix */
          ck = cc +o*c;         /* and (sorted) cluster index matrix */
          r  = (db[z]*2+dk[ck[z]]) /* get radius of ball a. cluster */
             * (1+4*DBL_EPSILON);  /* (account for rounding errors) */
          db[0] = INFINITY;     /* reinitialize the cluster heap */
          #if SORTED==1         /* if to also use sorted lists */
          if (z <= THRESH) {    /* if only few closest clusters */
            for (o = z, j = 0; (++j < c) && (dk[k = ck[j]] <= r); ) {
              d = vec_dist(mat_row(ctrs, k), vec, m);
              COUNTDC(1);       /* compute distance to next cluster */
              if (o > 0) o -= 1;/* get (next) position in heap */
              if (d < db[z]) {  /* if closer than old closest */
                for (p = o; ++p < z; ) {
                  db[p-1] = db[p]; cb[p-1] = cb[p]; }
                db[z-1] = db[z]; db[z] = d;
                cb[z-1] = ci[i]; ci[i] = k; }
              else if (d < db[0]) { /* if only some closest cluster */
                for (p = o; d < db[++p]; ) {
                  db[p-1] = db[p]; cb[p-1] = cb[p]; }
                db[p-1] = d; cb[p-1] = k;
              }                 /* update distances of old closest */
            } }                 /* and bring them into desc. order */
          else {                /* if many closest centers */
          #endif
            for (o = z, j = 0; (++j < c) && (dk[k = ck[j]] <= r); ) {
              d = vec_dist(mat_row(ctrs, k), vec, m);
              COUNTDC(1);       /* compute distance to next cluster */
              if (o > 0) o -= 1;/* get (next) position in heap */
              if (d < db[z]) {  /* if closer than closest cluster */
                xsift(db, cb, o, z, db[z], ci[i]);
                db[z] = d; ci[i] = k; }
              else if (d < db[0]) { /* if closer than some other */
                xsift(db, cb, o, z, d, k);
              }                 /* let second closest distance and */
            }                   /* cluster index sift down in heap */
          #if SORTED==1
          }  /* if (z <= THRESH) .. else .. */
          #endif
        }  /* if (mod == CLS_C2CALL) .. */
        else {                  /* if to recompute all distances */
          /* This is the standard generalized version:           */
          /* all data point to cluster distances are recomputed. */
          db[0] = INFINITY;     /* reinitialize the cluster heap */
          j = z;                /* start of already built heap part */
          #if SORTED==1
          if (z <= THRESH) {    /* if only few closest clusters */
            for (k = 0; k < c; k++) { /* traverse cluster centers, */
              if (k == o) continue;   /* but skip (old) closest */
              d = vec_dist(mat_row(ctrs, k), vec, m);
              COUNTDC(1);       /* compute distance to next cluster */
              if (j > 0) j -= 1;/* get (next) position in heap */
              if (d < db[z]) {  /* if closer than old closest */
                for (p = j; ++p < z; ) {
                  db[p-1] = db[p]; cb[p-1] = cb[p]; }
                db[z-1] = db[z]; db[z] = d;
                cb[z-1] = ci[i]; ci[i] = k; }
              else if (d < db[0]) { /* if only some closest cluster */
                for (p = j; d < db[++p]; ) {
                  db[p-1] = db[p]; cb[p-1] = cb[p]; }
                db[p-1] = d; cb[p-1] = k;
              }                 /* update distances of old closest */
            } }                 /* and bring them into desc. order */
          else {                /* if many closest centers */
          #endif
            for (k = 0; k < c; k++) { /* traverse cluster centers, */
              if (k == o) continue;   /* but skip (old) closest */
              d = vec_dist(mat_row(ctrs, k), vec, m);
              COUNTDC(1);       /* compute distance to next cluster */
              if (j > 0) j -= 1;/* get (next) position in heap */
              if (d < db[z]) {  /* if closer than closest cluster */
                xsift(db, cb, j, z, db[z], ci[i]);
                db[z] = d; ci[i] = k; }
              else if (d < db[0]) {  /* if closer than some other */
                xsift(db, cb, j, z, d, k);
              }                 /* let the second closest distance */
            }                   /* and corresponding cluster index */
          #if SORTED==1
          }                     /* sift down in (maximum) heap */
          #endif
        }  /* if (mod == CLS_C2CALL) .. else .. */
      }  /* for (i = 0; i < n; i++) .. */
      dmax = update(ctrs, cagg, data, ci);
      COUNTDC(c);               /* update the cluster centers and */
      SWAP(cagg, ctrs, mat);    /* exchange the center matrices */
    } }                         /* (current centers and aggregation) */

  /* --- z group bounds: yinyang approach [Ding et al. 2015] --- */
  else if (opt == CLS_YINYANG){ /* method from [Ding et al. 2015] */
    dsts = dx = (double*)malloc((size_t)(n*z+n+c+z) *sizeof(double)
                               +(size_t)(n+c+z+1)   *sizeof(DIMID));
    if (!dsts) { mat_delete(cagg); return -1; }
    dl = dx +n*z +n;            /* allocate arrays for distances */
    dc = dl +c;                 /* to closest clusters centers, */
    ci = (DIMID*)(dc +z);       /* for closest cluster indices for */
    cg = ci +n;                 /* data points, and group indices */
    cx = cg +c +1;              /* and cluster group offsets */
    mat_rallwgt(ctrs, 1.0);     /* weight all cluster centers equally */
    mat = mat_createx(z, m, MAT_ROWWGTS);
    if (!mat                    /* cluster the cluster centers */
    || (cls_ctrsel(mat, ctrs, CLS_KMPP,1,0, outids,outcnt, drand) < 0)
    || (cls_xhcm(mat, ctrs, CLS_ELKAN,1000, outids,outcnt, 0, cg) < 0)){
      free(dsts); mat_delete(cagg); return -1; }
    mat_sorti(ctrs, +1, cg);    /* sort centers by group index */
    dim_qsort(cg, (size_t)c, +1);
    memset(cx-1, 0, (size_t)(z+1) *sizeof(DIMID));
    for (i = 0; i < c; i++)     /* count the number of clusters */
      cx[cg[i]] += 1;           /* in each cluster group */
    for (j = 0; j < z; j++)     /* compute the offsets to the */
      cx[j] += cx[j-1];         /* starts of each cluster group */

    #ifdef VERBOSE              /* if verbose progress reporting */
    fprintf(stderr, "       0\b\b\b\b\b\b\b\b");
    #endif                      /* (first epoch is special) */
    db = dx;                    /* epoch 0: traverse the data points */
    for (i = 0; i < n; db += z+1, i++) {
      vec = mat_row(data, i);   /* get the next data point */
      dl[ci[i] = 0] = d = vec_dist(mat_row(ctrs, 0), vec, m);
      for (k = 0; ++k < c; ) {  /* traverse the clusters */
        dl[k] = vec_dist(mat_row(ctrs, k), vec, m);
        if (dl[k] < d) d = dl[ci[i] = k];
      }                         /* find distance to closest cluster */
      db[z] = d;                /* and the index of this cluster */
      for (k = 0; k < c; k++)   /* traverse the clusters again */
        if ((k != ci[i]) && (dl[k] < db[cg[k]]))
          db[cg[k]] = dl[k];    /* find minimum distance per group */
    }
    COUNTDC(n*c);               /* count distance computations */
    dmax = update(ctrs, cagg, data, ci);
    COUNTDC(c);                 /* update the cluster centers and */
    SWAP(cagg, ctrs, mat);      /* exchange the center matrices */
    if (!(mod & CLS_C2CDIST))   /* if no min. cls. to cls. distances */
      memset(dl, 0, (size_t)c *sizeof(double));

    for (e = 0; (++e < epochs) && (dmax > 0); ) {
      #ifdef VERBOSE            /* if verbose progress reporting */
      if ((e & 0x00ff) == 0)    /* print the current number of epochs */
        fprintf(stderr, "%8"DIMID_FMT"\b\b\b\b\b\b\b\b", e);
      #endif
      for (j = 0; j < z; j++) { /* traverse the cluster groups */
        for (dc[j] = 0, k = cx[j-1]; k < cx[j]; k++) {
          d = mat_rgetwgt(ctrs, k);
          if (d > dc[j]) dc[j] = d;
        }                       /* for each cluster group: determine */
      }                         /* the maximum movement of a cluster */
      if (mod & CLS_C2CDIST) {  /* if cluster to cluster distances */
        c2cdist(ctrs, dl, 0);   /* for each cluster: */
        COUNTDC(c*(c-1)/2);     /* compute (half) the distance */
      }                         /* to the closest other cluster */
      db = dx;                  /* traverse the data points */
      for (i = 0; i < n; db += z+1, i++) {
        o = ci[i];              /* update bound for closest cluster */
        db[z] += mat_rgetwgt(ctrs, o);
        for (d = INFINITY, j = z; --j >= 0; ) {
          db[j] -= dc[j];       /* update distance bounds for groups */
          if (db[j] < d) d = db[j];   /* find bound for distance */
        }                       /* to the second closest cluster */
        if (d < dl[o])          /* if half the distance to closest */
          d = dl[o];            /* other cluster is a better bound */
        if (db[z] <= d)         /* if closest cluster is the same, */
          continue;             /* no distance comps. are needed */
        vec = mat_row(data, i); /* get the data point to process */
        db[z] = r = vec_dist(mat_row(ctrs, o), vec, m);
        COUNTDC(1);             /* make upper distance bound tight */
        if (db[z] <= d)         /* if closest cluster is the same, */
          continue;             /* no distance comps. are needed */
        for (j = 0; j < z; j++){/* traverse the cluster groups, */
          if (db[z] <= db[j])   /* but skip those cluster groups that */
            continue;           /* cannot contain the closest cluster */
          db[j] = INFINITY;     /* init. the group distance bound */
          for (k = cx[j-1]; k < cx[j]; k++) {
            if (k == o) d = r;  /* traverse clusters in group */
            else      { d = vec_dist(mat_row(ctrs, k), vec, m);
              COUNTDC(1); }     /* compute distance to next cluster */
            if (d < db[z]) {    /* if new closest cluster found */
              if (db[z] < db[cg[ci[i]]])  /* update group bound with */
                db[cg[ci[i]]] = db[z];    /* old closest distance */
              db[z] = d;        /* note distance to closest cluster */
              ci[i] = k; }      /* and index of closest cluster */
            else if (d < db[j]) /* if closer than currently closest */
              db[j] = d;        /* cluster in the j-th cluster group, */
          }                     /* update the cluster group bound */
        }  /* for (j = 0; j < z; j++) .. */
      }  /* for (i = 0; i < n; i++) .. */
      dmax = update(ctrs, cagg, data, ci);
      COUNTDC(c);               /* update the cluster centers and */
      SWAP(cagg, ctrs, mat);    /* exchange the center matrices */
    } }                         /* (current centers and aggregation) */

  /* --- no distance bounds --- */
  else {                        /* standard method (full computation) */
    dsts = db = (double*)malloc((size_t)n *sizeof(double)
                               +(size_t)n *sizeof(DIMID));
    if (!dsts) { mat_delete(cagg); return -1; }
    ci = (DIMID*)(db +n);       /* allocate data point information */
    for (dmax = 1.0, e = 0; (e < epochs) && (dmax > 0); e++) {
      #ifdef VERBOSE            /* if verbose progress reporting */
      if ((e & 0x00ff) == 0)    /* print the current number of epochs */
        fprintf(stderr, "%8"DIMID_FMT"\b\b\b\b\b\b\b\b", e);
      #endif
      for (i = 0; i < n; i++) { /* traverse the data points */
        vec = mat_row(data, i); /* get the data point to process */
        ci[i] = 0; db[i] = INFINITY;
        for (k = 0; k < c; k++){/* traverse the cluster centers */
          d = vec_dist(mat_row(ctrs, k), vec, m);
          if (d < db[i]) { db[i] = d; ci[i] = k; }
        }                       /* compute distance to next cluster */
      }                         /* and update closest cluster */
      COUNTDC(n*c);             /* count distance computations */
      dmax = update(ctrs, cagg, data, ci);
      COUNTDC(c);               /* update the cluster centers and */
      SWAP(cagg, ctrs, mat);    /* exchange the center matrices */
    }                           /* (current centers and aggregation) */
  }
  #ifdef BENCH                  /* if benchmark version */
  fprintf(stderr, "dc: %zd ", distcnt);
  #endif                        /* print number of distance comps. */
  if (cids)                     /* copy the data point assignments */
    memcpy(cids, ci, (size_t)n *sizeof(DIMID));
  if (e & 1) {                  /* if number of epochs is odd */
    mat = cagg; cagg = ctrs; ctrs = mat;
    mat_copy(ctrs, cagg);       /* exchange matrices to restore state */
  }                             /* and copy centers to input matrix */
  free(dsts);                   /* delete the work memory and */
  mat_delete(cagg);             /* the center aggregation matrix */
  mat_rallwgt(ctrs, 1.0);       /* reset the cluster center weights */
  return e;                     /* return the number of epochs */
}  /* cls_xhcm() */

/*----------------------------------------------------------------------
  References for bounds-based exact hard c-means:
  [Elkan 2003]
    Charles Elkan.
    Using the Triangle Inequality to Accelerate K-Means.
    Proc. 20th Int. Conf. on Machine Learning
    (ICML 2003, Washington, DC), 147-153.
    AAAI Press, Menlo Park, CA, USA 2003
  [Hamerly 2010]
    Greg Hamerly.
    Making K-Means Even Faster.
    SIAM Int. Conf. on Data Mining (SDM 2010, Columbus, OH), 130–140.
    Society for Industrial and Applied Mathematics,
    Philadelphia, PA, USA 2010
  [Drake 2013] (Annular algorithm)
    Jonathan Drake.
    Faster k-Means Clustering,
    Master's Thesis, Baylor University, Waco, TX, USA 2013
  [Ding et al. 2015] (YinYang algorithm)
    Yufei Ding, Yue Zhao, Yipeng Shen, Madanlal Musuvathi,
    and Todd Mytkowicz.
    YinYang k-means: A Drop-in Replacement of the Classic k-Means
    with Consistent Speedup.
    Proc. 32nd Int. Conf. on Machine Learning,
    (ICML 2015, Lille, France), 579–587,
    http://jmlr.org/proceedings/papers/v37/ding15.html
  [Newling and Fleuret 2016]
    James Newling and Francois Fleuret
    Fast K-Means with Accurate Bounds.
    Proc. 33rd Int. Conf. on Machine Learning
    (ICML 2016, New York, NY), 936-944.
    AAAI Press, Menlo Park, CA, USA 2016
----------------------------------------------------------------------*/

double cls_hobjfn (MATRIX *ctrs, MATRIX *data)
{                               /* --- compute objective function */
  DIMID  n, i;                  /* number of data points, point index */
  DIMID  c, k;                  /* number of clusters, cluster index */
  DIMID  m;                     /* number of dimensions/vector elems. */
  double d, dmin;               /* (minimum) squared distance to cls. */
  double e;                     /* value of objective function */
  const double *vec = NULL;     /* current data point */

  assert(ctrs && data           /* check the function arguments */
  &&    (mat_colcnt(ctrs) == mat_colcnt(data)));
  n = mat_rowcnt(data);         /* get the number of data points */
  c = mat_rowcnt(ctrs);         /* and the number of clusters */
  m = mat_colcnt(ctrs);         /* and the number of dimensions */
  e = 0.0;                      /* init. the objective function value */
  for (i = 0; i < n; i++) {     /* traverse the data points */
    vec = mat_row(data, i);     /* get the next data point */
    dmin = INFINITY;            /* init. minimum distance (index) */
    for (k = 0; k < c; k++) {   /* traverse the clusters */
      d = vec_sqrdist(mat_row(ctrs, k), vec, m);
      if (d < dmin) dmin = d;   /* find distance of data point */
    }                           /* to closest cluster center */
    e += dmin *mat_rgetwgt(data, i);
  }                             /* sum (squared) distances */
  return e;                     /* return objective function value */
}  /* cls_hobjfn() */
